import React, {useCallback, useMemo, useState} from "react";
import {Card} from "../Card";
import {
	shift,
	useClick,
	useFloating,
	useFocus,
	useHover,
	useInteractions, useRole, useTransitionStyles
} from "@floating-ui/react";
import {UseFloatingOptions} from "@floating-ui/react/dist/floating-ui.react";
import {mergeRefs} from "react-merge-refs";

/**
 * Fully managed floating content function.
 */
export type Managed<T = React.ReactElement|React.ReactNode> = (show: () => void, hide: () => void) => T;

/**
 * Allowed floating modes.
 */
export type FloatingMode = "always"|"click"|"hover"|"focus"|"managed";

/**
 * Role of a floating element.
 */
export type FloatRole = "tooltip" | "dialog" | "alertdialog" | "menu" | "listbox" | "grid" | "tree" | "select" | "label" | "combobox";

/**
 * A component to show something floating next to an element.
 */
export function Float({children, content, className, mode, role, floatingOptions}: {
	children: (React.ReactElement & React.ClassAttributes<HTMLElement>)|Managed<(React.ReactElement & React.ClassAttributes<HTMLElement>)>;
	content?: React.ReactNode|Managed<React.ReactNode>;
	className?: string;
	mode?: FloatingMode;
	role?: FloatRole;
	floatingOptions?: UseFloatingOptions;
}): React.ReactElement
{
	// By default, use "always" mode.
	if (!mode) mode = "always";

	// Followed show status.
	const [shown, setShown] = useState(false);

	// Create show / hide functions.
	const show = useCallback(() => setShown(true), [setShown]);
	const hide = useCallback(() => setShown(false), [setShown]);

	// If show mode is "always", always show the floating part after render.
	if (mode == "always")
	{
		setTimeout(() => {
			setShown(true);
		}, 0);
	}

	// Floating initialization.
	const { refs, floatingStyles, context } = useFloating(
		useMemo(() => (Object.assign({
			open: shown,
			onOpenChange: setShown,
			middleware: [shift()],
		} as UseFloatingOptions, floatingOptions)), [floatingOptions, shown, setShown])
	);

	// Interactions initialization.
	const hover = useHover(context, useMemo(() => ({ enabled: mode == "hover" }), [mode]));
	const focus = useFocus(context, useMemo(() => ({ enabled: mode == "focus", visibleOnly: false }), [mode]));
	const click = useClick(context, useMemo(() => ({ enabled: mode == "click" }), [mode]));
	const roleProps = useRole(context, {
		role: role,
		enabled: !!role,
	});
	const {getReferenceProps, getFloatingProps} = useInteractions([roleProps, hover, focus, click]);

	// Transition configuration.
	const {isMounted, styles: transitionStyles} = useTransitionStyles(context, {
		duration: 200,
		common: ({side}) => ({
			transformOrigin: {
				top: "bottom",
				bottom: "top",
				left: "right",
				right: "left",
			}[side],
		}),
		initial: {
			transform: "scale(0.66)",
			opacity: "0",
		},
	});

	// Change the child element to use the reference and the interactions properties.
	const referencedChild = useMemo(() => {
		// Render the children if a managed floating function is passed.
		const child = typeof children == "function" ? children(show, hide) : children;

		return React.cloneElement(child,
			Object.assign(
				{
					// Pass references.
					ref: mergeRefs([refs.setReference, child?.ref]),
				},
				// Get interaction properties.
				getReferenceProps(),
			),
		);
	}, [children, show, hide, refs.setReference, getReferenceProps]);

	// Update floating content.
	const floatingContent = useMemo(() => (
		// Render the children if a managed floating function is passed.
		typeof content == "function" ? content(show, hide) : content
	), [shown, show, hide, content]);

	return (
		<>
			{referencedChild}

			{ // Showing floating element if the state says to do so.
				isMounted &&
				<div ref={refs.setFloating} style={floatingStyles} {...getFloatingProps()} className={"floating"}>
					<Card style={transitionStyles} className={`floating${className ? ` ${className}` : ""}`}>
						{floatingContent}
					</Card>
				</div>
			}
		</>
	);
}
